#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "tools.h"

#define MAX(x,y) ((x)>(y)?(x):(y))

#ifdef FEDOROV_DEBUG_
	#include <stdio.h>
	#define mexPrintf printf
	#define mexEvalString printf
#else
	#include "mex.h"
#endif

void read_data(FILE* f, double* A, int len) {

	int i;
	double t;

	for(i = 0; i < len; i++) {
		fscanf(f, "%lf\n", &t);
		A[i] = t;
	}

}

void print_matrix(double* A, int n, int m) {

	int i, j;

	for(i = 0; i < n; i++) {
		for(j = 0; j < m; j++)
			printf("%lf ", A[i*m+j]);
		printf("\n");
	}

}

void update(double* A, double* invA, double* logdetA, double* X, int p, int id_out, int id_in, double* workspace) {

	double* x_out = X + id_out * p;
	double* x_in = X + id_in * p;

	rank1_update(A, invA, logdetA, x_in, 1.0, p, workspace);
	rank1_update(A, invA, logdetA, x_out, -1.0, p, workspace);

}

/* fedorov(objtype, X, ret, A, invA, logdetA, cov0, withrep, verbose, n, p, k, &iters, maxT) */
void fedorov(char objtype, double* X, int* inds, double* A, double* invA, double* logdetA, double* cov0, int withrep, int verbose, int n, int p, int k, int* iters, int maxT) {

	int* flags = (int*)(malloc(n * sizeof(int)));

	double* workspace = (double*)(malloc(2*p * sizeof(double)));
	
	int i, j;
	double f_opt, f_now;
	int i_opt, j_opt;
	int iter = 0;

	f_opt = compute_objective(A, invA, *logdetA, p, objtype, cov0, X, n, workspace);
	if (verbose) {
		mexPrintf("f_init: %lf\n", f_opt);
	}

	for(i = 0; i < k; i++)
		flags[inds[i]] = 1;

	/* Fedorov's exchange algorithm */
	while (iter < maxT) {

		iter ++;
		if (verbose) {
			mexPrintf("iter = %d: ", iter);
		}

		i_opt = -1;
		j_opt = -1;
		for(i = 0; i < k; i++)
			for(j = 0; j < n; j++)
				if (inds[i] != j && (withrep || !flags[j])) {
					update(A, invA, logdetA, X, p, inds[i], j, workspace);
					f_now = compute_objective(A, invA, *logdetA, p, objtype, cov0, X, n, workspace);
					update(A, invA, logdetA, X, p, j, inds[i], workspace);
					if (f_now < f_opt) {
						f_opt = f_now;
						i_opt = i;
						j_opt = j;
					}
				}
		if (i_opt == -1) {
			mexPrintf("Done!\n");
			mexEvalString("pause(.001);");
			break;
		}

		update(A, invA, logdetA, X, p, inds[i_opt], j_opt, workspace);
		flags[inds[i_opt]] = 0;
		flags[j_opt] = 1;
		inds[i_opt] = j_opt;
		if (verbose) {
			mexPrintf("fopt = %lf\n", f_opt);
			mexEvalString("pause(.001);");
		}


	}

	if (iters) *iters = iter;

	free(flags);
	free(workspace);

}


/* objtype, X, c0, A, invA, logdetA, cov0, withrep/worep, verbose 
c0 is a 1 x k vector, corresponding to an initial set
A = X_c0 * X_c0^T
return: either 1 or 2 variables; C and # of iterations
*/
#ifndef FEDOROV_DEBUG_
void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray *prhs[]) {

	char* objtype = (char*)mxGetChars(prhs[0]);
	mexPrintf("objtype = %c\n", objtype[0]);
	mexEvalString("pause(.001);");
	double* X = mxGetPr(prhs[1]);
	double* C = mxGetPr(prhs[2]);
	double* A = mxGetPr(prhs[3]);
	double* invA = mxGetPr(prhs[4]);
	double logdetA = mxGetScalar(prhs[5]);
	double* cov0 = mxGetPr(prhs[6]);
	int withrep = (int)(mxGetScalar(prhs[7]));
	int verbose = (int)(mxGetScalar(prhs[8]));
	int maxT = (int)(mxGetScalar(prhs[9]));

	int p = (int)(mxGetM(prhs[1]));
	int n = (int)(mxGetN(prhs[1]));
	int k = MAX((int)(mxGetM(prhs[2])), (int)(mxGetN(prhs[2])));
	int iters = 0;

	int* ret;
	int i, j;

	plhs[0] = mxCreateNumericMatrix(1, k, mxINT32_CLASS, mxREAL);
	ret = (int*)mxGetData(plhs[0]);
	for(i = 0; i < k; i++)
		ret[i] = (int)(C[i]) - 1;

	fedorov(objtype[0], X, ret, A, invA, &logdetA, cov0, withrep, verbose, n, p, k, &iters, maxT);
	
	for(i = 0; i < k; i++) {
		ret[i] = ret[i] + 1;
	}
	if (nlhs > 1) {
		plhs[1] = mxCreateDoubleScalar(iters);
	}

}
#endif

#ifdef FEDOROV_DEBUG_
int main() {

	int n = 1000;
	int p = 25;
	int k = 100;

	double* X = (double*)(malloc(n*p * sizeof(double)));
	double* C = (double*)(malloc(k * sizeof(double)));
	double* A = (double*)(malloc(p*p * sizeof(double)));
	double* invA = (double*)(malloc(p*p * sizeof(double)));
	double logdetA;
	double* cov0 = (double*)(malloc(p*p * sizeof(double)));

	int* inds = (int*)(malloc(k * sizeof(int)));

	FILE* f = fopen("input.txt", "r");

	int i;
	int iters = 0;

	read_data(f, X, n*p);
	read_data(f, C, k);
	read_data(f, A, p*p);
	read_data(f, invA, p*p);
	read_data(f, &logdetA, 1);
	read_data(f, cov0, p*p);

	fclose(f);

	for(i = 0; i < k; i++)
		inds[i] = (int)(C[i]) - 1;

	fedorov('E', X, inds, A, invA, &logdetA, cov0, 0, 1, n, p, k, &iters, maxT);

	for(i = 0; i < k; i++)
		printf("%d ", inds[i] + 1);
	printf("\n");

	return 0;

}
#endif