#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "tools.h"

#define MAX(x,y) ((x)>(y)?(x):(y))
#define MIN(x,y) ((x)<(y)?(x):(y))

#ifdef PIPAGE_DEBUG_
	#include <stdio.h>
	#define mexPrintf printf
	#define mexEvalString printf
#else
	#include "mex.h"
#endif

void read_data(FILE* f, double* A, int len) {

	int i;
	double t;

	for(i = 0; i < len; i++) {
		fscanf(f, "%lf\n", &t);
		A[i] = t;
	}

}

void print_matrix(double* A, int n, int m) {

	int i, j;

	for(i = 0; i < n; i++) {
		for(j = 0; j < m; j++)
			printf("%lf ", A[i*m+j]);
		printf("\n");
	}

}

/* pipage(objtype, ret, X, A, invA, &logdetA, cov0, n, p, k, w0, verbose_step) */
void pipage(char objtype, int* ret, double* X, double* A, double* invA, double* logdetA, double* cov0, int n, int p, int k, double* w, int verbose_step) {

	double* workspace = (double*)(malloc(2*p * sizeof(double)));

	int i, j, iter;
	double f_opt = compute_objective(A, invA, *logdetA, p, objtype, cov0, X, n, workspace);
	double f_now;
	int i_opt, j_opt;
	double delta, delta_opt;

	int nonzeros = 0;
	for(i = 0; i < n; i++) {
		if (fabs(w[i]) > 1e-9) nonzeros ++;
	}

	/* remove one data point at a time */
	iter = 0;
	while (nonzeros > k && iter <= 200) {

		f_opt = 1e100;
		i_opt = -1; j_opt = -1; delta_opt = 0;
		for(i = 0; i < n; i++)
			if (fabs(w[i]) > 1e-9) {
				for(j = 0; j < n; j++) {
					if (i != j && fabs(w[j]-1) > 1e-9) {
						delta = MIN(w[i], 1-w[j]);
						rank1_update(A, invA, logdetA, X + j * p, delta, p, workspace);
						int success = rank1_update(A, invA, logdetA, X + i * p, -delta, p, workspace);						
						if (!success) {
							rank1_update(A, invA, logdetA, X + j * p, -delta, p, workspace);
							continue;
						}
						f_now = compute_objective(A, invA, *logdetA, p, objtype, cov0, X, n, workspace);
						rank1_update(A, invA, logdetA, X + i * p, delta, p, workspace);
						rank1_update(A, invA, logdetA, X + j * p, -delta, p, workspace);
						if (f_now < f_opt) {
							f_opt = f_now;
							i_opt = i; j_opt = j;
							delta_opt = delta;
						}
					}
				}
			}

		/* mexPrintf("i_opt = %d, j_opt = %d, delta_opt = %lf, f_opt = %lf\n", i_opt, j_opt, delta_opt, f_opt); */
		if (i_opt == -1) break;

		rank1_update(A, invA, logdetA, X + j_opt * p, delta_opt, p, workspace);
		rank1_update(A, invA, logdetA, X + i_opt * p, -delta_opt, p, workspace);
		w[i_opt] -= delta_opt;
		w[j_opt] += delta_opt;
		if (fabs(w[i_opt]) < 1e-9) nonzeros --;

		if (verbose_step > 0 && iter % verbose_step == 0) {
			mexPrintf("Iter %d: f_opt = %lf, nonzeros = %d\n", iter, f_opt, nonzeros);
			mexEvalString("pause(.001);");
		}

		iter ++;

	}

	/* record the final solution */
	j = 0;
	for(i = 0; i < n; i++)
		if (fabs(w[i]) > 1e-9) {
			ret[j++] = i;
			if (j == k) break;
	}

	/* release allocated memory */
	free(workspace);

}


/* objtype, X, c0, A, invA, logdetA, cov0, w0, verbose 
c0 is a 1 x k vector, corresponding to an initial set
A = X_c0 * X_c0^T
return: either 1 or 2 variables; C and # of iterations
*/
#ifndef PIPAGE_DEBUG_
void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray *prhs[]) {

	char* objtype = (char*)mxGetChars(prhs[0]);
	mexPrintf("objtype = %c\n", objtype[0]);
	mexEvalString("pause(.001);");
	double* X = mxGetPr(prhs[1]);
	int k = (int)(mxGetScalar(prhs[2]));
	double* A = mxGetPr(prhs[3]);
	double* invA = mxGetPr(prhs[4]);
	double logdetA = mxGetScalar(prhs[5]);
	double* cov0 = mxGetPr(prhs[6]);
	double* w0 = mxGetPr(prhs[7]);
	int verbose_step = (int)(mxGetScalar(prhs[8]));

	int p = (int)(mxGetM(prhs[1]));
	int n = (int)(mxGetN(prhs[1]));
	int iters = 0;

	int* ret;
	int i, j;

	plhs[0] = mxCreateNumericMatrix(1, k, mxINT32_CLASS, mxREAL);
	ret = (int*)mxGetData(plhs[0]);

	pipage(objtype[0], ret, X, A, invA, &logdetA, cov0, n, p, k, w0, verbose_step);
	
	for(i = 0; i < k; i++) {
		ret[i] = ret[i] + 1;
	}
	
}
#endif
