function [ w, tot_iters, f_opt ] = solve_pi_pgd( X, alpha, beta, eps, infty_norm, objtype, lambda, maxT, verbose_step )
% Solve the optimal design problem using PGD via backtracking line search
% Input:
%   - X: n x p, the design pool matrix
%   - 0<alpha<0.5, 0<beta<1, t0: parameters for Armijo's rule
%   - eps: terminate if relative difference of the objective < eps
%   - maxT: maximum number of gradient descent iterations
%   - infty_norm: infinity norm bound
%   - lambda: used in smoothed (Ridge) regression
%   - objtype: type of objectives; {'A', 'D', 'T', 'E', 'G', 'I'}
%   - verbose_step: (optioal) number of steps to output objective value;
%   set to 0 for the silence mode. default 1.
% Output:
%   - w: n x 1, the subsampling weight vector, normalized to sum 1.
%   - tot_iters: integer, total number of iterations

    if (nargin < 9)
        verbose_step = 1;
    end
    if (nargin < 8)
        maxT = 1000;
    end
    if (nargin < 7)
        lambda = 0;
    end
    
    [n, p] = size(X);
    assert(0 < alpha && alpha <= 0.5);
    assert(0 < beta && beta < 1);
    
    % initialization
    w = 1/n * ones(n, 1);
    %w = rand(n, 1);
    %w = w / sum(w);
    
    Sigma = X' * diag(w) * X;
    f_opt = compute_objective(X, Sigma, objtype, lambda);
    if (objtype == 'd' || objtype == 'D'); f_opt = exp(f_opt); end
    fprintf('Start %c-optimality: f_0 = %f\n', objtype, f_opt);
    if (objtype == 'd' || objtype == 'D'); f_opt = log(f_opt); end
    
    % gradient descent
    last_fopt = 1e100;
    iter = 0;
    while (iter < maxT && abs(f_opt - last_fopt) / abs(last_fopt) > eps)
        
        % compute the gradient
        grad = compute_gradient(X, Sigma, objtype, lambda); % 1 x n
        
        % compute stepsize using backtracking line search
        t = 1 / norm(grad);
        while (t > 0)
            wt = project_L1LInf(w-t*grad, infty_norm, 1, ones(n, 1));
            if (sum(wt > 0) < p)
                t = t * beta;
                continue;
            end
            if (min(wt) < 0)
                t = t * beta;
                continue;
            end
            new_fopt = compute_objective(X, X'*diag(wt)*X, objtype, lambda);
            if (new_fopt - f_opt <= alpha * (grad' * (wt - w)))
                break;
            end
            t = t * beta;
            
        end
        
        %projective gradient descent
        w = project_L1LInf(w-t*grad, infty_norm, 1, ones(n, 1));
        Sigma = X'*diag(w)*X;
        last_fopt = f_opt;
        f_opt = compute_objective(X, Sigma, objtype, lambda);
        
        % incraese $t$ and output current objective, if applicable
        iter = iter + 1;
        if (mod(iter, verbose_step) == 0)
            if (objtype == 'd' || objtype == 'D'); f_opt = exp(f_opt); end
            fprintf('Step %d: objective = %f\n', iter, f_opt);
            if (objtype == 'd' || objtype == 'D'); f_opt = log(f_opt); end;
        end
        
    end
    
    fprintf('Completed. # of nonzero weights = %d\n', sum(w > zeros(n, 1)));
    fprintf('Total # of iterations: %d\n', iter);
    tot_iters = iter;

end

