#include <math.h>
#include <stdlib.h>

#include "mex.h"

#define IND(i,j,d) ((i)*d+(j))
#define SQR(x) ((x)*(x))

double spnorm(double* A, int d, double* workspace) {

	double eps = 1e-9;
	double last_f = 1;
	double f = 0;
	int i, j;
	double multiplier = 0;

	double* x = workspace;
	double* y = workspace + d;

	/* initialization */
	for(i = 0; i < d; i++)
		x[i] = 1.0 / sqrt(d);

	/* power update: y = normalize(Ax) */
	while (fabs(last_f - f) > eps && fabs((last_f - f) / f) > eps) {
		last_f = f;
		for(i = 0; i < d; i++) {
			y[i] = 0;
			for(j = 0; j < d; j++)
				y[i] += A[IND(i,j,d)] * x[j];
		}
		f = 0;
		for(i = 0; i < d; i++)
			f += SQR(y[i]);
		f = f / sqrt(f);
		multiplier = 1.0 / f;
		for(i = 0; i < d; i++)
			x[i] = y[i] * multiplier;
	}

	return f;

}

int rank1_update(double* A, double* invA, double* logdetA, double* x, double lambda, int d, double* workspace) {

	double t = 0;
	double multiplier = 0;
	int i, j;

	double* z = workspace;	

	/* update logdetA */
	t = 0;
	for(i = 0; i < d; i++)
		for(j = 0; j < d; j++)
			t += lambda * invA[IND(i,j,d)] * x[i] * x[j];
	if (1+t <= 1e-9) return 0;
	if (t < 1)
		*logdetA += log1p(t);
	else
		*logdetA += log(1+t);

	/* update invA */
	for(i = 0; i < d; i++) {
		z[i] = 0;
		for(j = 0; j < d; j++)
			z[i] += invA[IND(i,j,d)] * x[j];
	}
	multiplier = 1.0 / (1 + t);
	for(i = 0; i < d; i++)
		for(j = 0; j < d; j++)
			invA[IND(i,j,d)] -= multiplier * lambda * z[i] * z[j];

	/* update A */
	for(i = 0; i < d; i++)
		for(j = 0; j < d; j++)
			A[IND(i,j,d)] += lambda * x[i] * x[j];

	return 1;

}

double compute_objective(double* A, double* invA, double logdetA, int d, char objtype, double* cov0, double* X, int n, double* workspace) {

	int i, j, ell;
	double ret = 0;

	switch(objtype) {
		case 'a': case 'A':
			for(i = 0; i < d; i++)
				ret += invA[IND(i,i,d)];
			return ret / d;
		case 'd': case 'D':
			return exp(-logdetA / d);
		case 't': case 'T':
			for(i = 0; i < d; i++)
				ret += A[IND(i,i,d)];
			return d / ret;
		case 'e': case 'E':
			return spnorm(invA, d, workspace);
		case 'i': case 'I':
			for(i = 0; i < SQR(d); i++)
				ret += invA[i] * cov0[i];
			return ret;
		case 'g': case 'G':
			for(ell = 0; ell < n; ell++) {
				double sum = 0;
				for(i = 0; i < d; i++)
					for(j = 0; j < d; j++)
						sum += invA[IND(i,j,d)] * X[IND(ell,i,d)] * X[IND(ell,j,d)];
				if (sum > ret) ret = sum;
			}
			return ret;
		default: 
			return -1;
	}

}
